
#toolbox expander {
  border-top: 1px solid @unfocused_borders;
  margin-top: 6px;
  padding-top: 6px;
  padding-bottom: 6px;
}

#toolbox expander:first-child {
  border-top: none;
  margin-top: 0;
}

#toolbox label {
  font-weight: bold;
}

#diagrams {
  border-top-width: 0;
}

diagramview {
  background: white;
}

.status-window box {
  padding: 12px;
}

.dropdown-small {
  padding-left: 4px;
  padding-right: 4px;
}

flowboxchild {
  padding: 8px;
  border-radius: 8px;
}

.greeter list row {
  border-bottom: 1px solid alpha(@borders, 0.2);
  transition-property: background, opacity;
}

.greeter list row:last-child {
  border-bottom: none;
}

.greeter frame border {
  border-color: alpha(@borders, 0.6);
}

.greeter flowboxchild {
  border-radius: 12px;
  outline-offset: -3px;
}

.greeter-title {
  font-weight: bold;
  padding: 18px 0 6px;
}

.info {
  border-radius: 16px;
  font-size: small;
  padding: 1px 6px;
}

.greeter preferencesgroup .title {
  font-weight: bold;
}

.greeter preferencesgroup .subtitle {
  font-weight: normal;
}

.greeter preferencesgroup row {
  padding: 6px 0 6px 6px;
  min-width: 400px;
}

box.propertypage > * {
  margin-top: 6px;
}

expander-widget.propertypage {
  margin-top: 12px;
}

.propertypage box.horizontal:not(.linked) {
  border-spacing: 6px;
}

.propertypage .title {
  font-weight: bold;
}

.propertypage .info {
  margin: 3px 2px 0 2px;
}

frame > textview {
  padding: 6px;
}

row {
  min-height: 24px;
}

.data-table checkbutton {
  padding: 0px;
}

.data-table box checkbutton {
  margin-right: 4px;
 }

.data-table .stereotype-value {
  padding-left: 24px;
 }

textfield text {
  background-color: @theme_bg_color;
  color: @theme_fg_color;
  border: 1px solid @theme_bg_color;
}

textfield text selection {
  background-color: alpha(@theme_selected_bg_color, 0.3);
}

.info:hover {
  background-color: @shade_color;
}

.info-popover {
  margin: 6px;
}

#model_browser treeexpander {
  margin: 0 6px;
}

#model_browser .row {
  border-top: 0 solid @shade_color;
  transition: border 200ms;
}

#model_browser .move-element-above {
  border-top-width: 12px;
}

#model_browser .move-element-above:drop(active) {
  box-shadow: none;
}

#element_editor {
  border-left: 1px solid @unfocused_borders;
}

.editor-wide {
  min-width: 350px;
}

#toolbox > viewport > box {
  margin: 6px;
}

flowboxchild:hover {
  background-color: @shade_color;
}

flowboxchild:active {
  background-color: alpha(@shade_color, 2);
}
